# -*- coding: iso-8859-15 -*-
# (c) Mikael Högqvist

from benri.app.wsgiapp import WSGIApp
from benri.app.wsgicollection import Collection as WSGICollection
from benri.app.json import json, JSON_CONTENT_TYPE
